-- by mor2000

--------------------
-- UPK_ObjectSpawner (spawns objects like pallets, bales, bigbags, basketballs, .. you name it)

local UPK_ObjectSpawner_mt = ClassUPK(UPK_ObjectSpawner, UniversalProcessKit)
InitObjectClass(UPK_ObjectSpawner, "UPK_ObjectSpawner")
UniversalProcessKit.addModule("objectspawner", UPK_ObjectSpawner)

UPK_ObjectSpawner.objectTypes = {}
UPK_ObjectSpawner.objectTypeToClass = {}

-- in-game pallets
UPK_ObjectSpawner.objectTypes["fertilizerTank"] = "$data/objects/pallets/fertilizerTank.i3d"
UPK_ObjectSpawner.objectTypeToClass["fertilizerTank"] = FillablePallet
UPK_ObjectSpawner.objectTypes["palletPoplar"] = "$data/objects/pallets/palletPoplar.i3d"
UPK_ObjectSpawner.objectTypeToClass["palletPoplar"] = FillablePallet
UPK_ObjectSpawner.objectTypes["palletPotato"] = "$data/objects/pallets/palletPotato.i3d"
UPK_ObjectSpawner.objectTypeToClass["palletPotato"] = FillablePallet
UPK_ObjectSpawner.objectTypes["palletSugarBeet"] = "$data/objects/pallets/palletSugarBett.i3d"
UPK_ObjectSpawner.objectTypeToClass["palletSugarBeet"] = FillablePallet
UPK_ObjectSpawner.objectTypes["treeSaplingPallet"] = "$data/objects/pallets/treeSaplingPallet.i3d"
UPK_ObjectSpawner.objectTypeToClass["treeSaplingPallet"] = FillablePallet
UPK_ObjectSpawner.objectTypes["woolPallet"] = "$data/objects/pallets/woolPallet.i3d"
UPK_ObjectSpawner.objectTypeToClass["woolPallet"] = FillablePallet
	
-- in-game bales
UPK_ObjectSpawner.objectTypes["roundbaleGrass"] = "$data/objects/roundbales/roundbaleGrass_w112_d130.i3d"
UPK_ObjectSpawner.objectTypeToClass["roundbaleGrass"] = Bale
UPK_ObjectSpawner.objectTypes["roundbaleHay"] = "$data/objects/roundbales/roundbaleHay_w112_d130.i3d"
UPK_ObjectSpawner.objectTypeToClass["roundbaleHay"] = Bale
UPK_ObjectSpawner.objectTypes["roundbaleSilage"] = "$data/objects/roundbales/roundbaleSilage_w112_d130.i3d"
UPK_ObjectSpawner.objectTypeToClass["roundbaleSilage"] = Bale
UPK_ObjectSpawner.objectTypes["roundbaleStraw"] = "$data/objects/roundbales/roundbaleStraw_w112_d130.i3d"
UPK_ObjectSpawner.objectTypeToClass["roundbaleStraw"] = Bale
UPK_ObjectSpawner.objectTypes["baleGrass"] = "$data/objects/squarebales/baleGrass240.i3d"
UPK_ObjectSpawner.objectTypeToClass["baleGrass"] = Bale
UPK_ObjectSpawner.objectTypes["baleHay"] = "$data/objects/squarebales/baleHay240.i3d"
UPK_ObjectSpawner.objectTypeToClass["baleHay"] = Bale
UPK_ObjectSpawner.objectTypes["baleStraw"] = "$data/objects/squarebales/baleStraw240.i3d"
UPK_ObjectSpawner.objectTypeToClass["baleStraw"] = Bale
	
-- in-game bigbags
UPK_ObjectSpawner.objectTypes["bigBagFertilizer"] = "$data/objects/bigBagContainer/bigBagContainerFertilizer.i3d"
UPK_ObjectSpawner.objectTypeToClass["bigBagFertilizer"] = FillablePallet
UPK_ObjectSpawner.objectTypes["bigBagPigFood"] = "$data/objects/bigBagContainer/bigBagContainerPigFood.i3d"
UPK_ObjectSpawner.objectTypeToClass["bigBagPigFood"] = FillablePallet
UPK_ObjectSpawner.objectTypes["bigBagSeeds"] = "$data/objects/bigBagContainer/bigBagContainerSeeds.i3d"
UPK_ObjectSpawner.objectTypeToClass["bigBagSeeds"] = FillablePallet
	
-- in-game objects
UPK_ObjectSpawner.objectTypes["basketball"] = "$data/objects/basketball/basketball.i3d"
UPK_ObjectSpawner.objectTypeToClass["basketball"] = Basketball
	
-- control for spawned and not spawned objects
UPK_ObjectSpawner.nextObjectId = 0
function UPK_ObjectSpawner.getNextObjectId()
	UPK_ObjectSpawner.nextObjectId = UPK_ObjectSpawner.nextObjectId + 1
	return UPK_ObjectSpawner.nextObjectId
end

function UPK_ObjectSpawner:new(nodeId, parent)
	printFn('UPK_ObjectSpawner:new(',nodeId,', ',parent,')')
	local self = UniversalProcessKit:new(nodeId, parent, UPK_ObjectSpawner_mt)
	registerObjectClassName(self, "UPK_ObjectSpawner")
	
	-- object
	
	-- just use attribute "object" with pre-set object types like "roundbaleGrass"
	-- or provide a filename in "object" and a class name in "objectClass" (default: Object)
	
	local objectStr = getStringFromUserAttribute(nodeId, "object", "")
	
	self.objectFilename = ""
	self.objectClass = nil
	
	if objectStr=="" then
		self:printErr('no object name given')
		return false
	elseif UPK_ObjectSpawner.objectTypes[objectStr]~=nil then
		-- preset object
		self:printInfo('preset object ',objectStr,' chosen')
		self.objectFilename = getLongFilename(UPK_ObjectSpawner.objectTypes[objectStr],nil)
		self.objectClass = UPK_ObjectSpawner.objectTypeToClass[objectStr]
	else
		-- filename
		local a,_=string.find(objectStr,"^[$%w_%/\\%.0-9]+%.i3d$")
		if a==nil then
			self:printErr('invalid object filename "'..tostring(objectStr)..'"')
			return false
		else
			-- object filename
			self.objectFilename = getLongFilename(objectStr,self.base.i18nNameSpace)
			if not fileExists(self.objectFilename) then
				self:printErr('couldnt find file "'..objectStr..'" ("'..objectFilename..'")')
				return false
			end
			
			-- object class
			local objectClassname = getStringFromUserAttribute(nodeId, "objectClass", "PhysicsObject")
			
			if _g[objectClassname]~=nil and type(_g[objectClassname])=="table" then
				self.objectClass = _g[objectClassname]
			elseif self.i18nNameSpace~=nil and _g[self.i18nNameSpace][objectClassname]~=nil and type(_g[self.i18nNameSpace][objectClassname])=="table" then
				self.objectClass = _g[self.i18nNameSpace][objectClassname]
			else
				self:printErr('no object class "'..objectClassname..'" found')
				return false
			end
			
			self:printInfo('ready to spawn "',objectStr,'"')
		end
	end
	
	if not (self.objectClass==PhysicsObject or derivesFromClass(self.objectClass,PhysicsObject)) then
		self:printErr('object to spawn is not an PhysicsObject')
		return false
	end
	
	self.isLimitedObjectTypeBale = (self.objectClass==Bale or derivesFromClass(self.objectClass,Bale))
	self.isLimitedObjectTypePallet = (self.objectClass==FillablePallet or derivesFromClass(self.objectClass,FillablePallet))
	
	-- spawn properties
	
	self.maxObjects = getNumberFromUserAttribute(nodeId, "maxObjects", 1, 1)
	self.spawnDelay = getNumberFromUserAttribute(nodeId, "spawnDelay", 1, 0.1)*1000 -- time in ms after trigger has less than max objects, then spawn again
	self.spawnPosition = getVectorFromUserAttribute(nodeId, "spawnPosition", "0 0 0")
	self.spawnRotation = getVectorFromUserAttribute(nodeId, "spawnRotation", "0 0 0")
	
	-- fillType, fillLevel
	
	-- fillType, fillLevel
	
	self.createFillType = getBoolFromUserAttribute(nodeId, "createFillType", false)
	self.objectCapacity = getNumberFromUserAttribute(nodeId, "capacity", nil, 0)
	self.objectFillLevel = getNumberFromUserAttribute(nodeId, "fillLevel", nil, 0, self.objectCapacity)
	local objectFillTypeStr = getStringFromUserAttribute(nodeId, "fillType")
	if objectFillTypeStr~=nil and UniversalProcessKit.fillTypeNameToInt[objectFillTypeStr]~=nil then
		self.objectFillType = UniversalProcessKit.fillTypeNameToInt[objectFillTypeStr]
	end
	
	-- prices
	
	self.moneyChangeId = getMoneyTypeId()
	
	 self.pricePerLiter = getNumberFromUserAttribute(nodeId, "pricePerLiter", 0)
	self.pricePerObject = getNumberFromUserAttribute(nodeId, "pricePerObject", 0)
	
	self.preferMapDefaultPrice = getBoolFromUserAttribute(nodeId, "preferMapDefaultPrice", false)
	self.pricePerLiterMultiplier = getVectorFromUserAttribute(nodeId, "pricePerLiterMultiplier", "1 1 1")
	self.pricesPerLiter = {}
	
	self.nrObjectsNotSold = 0

	-- statName
	
	self.statName = getStatNameFromUserAttribute(nodeId)
	
	-- actions
	
	self:getActionUserAttributes('OnObjectSpawned')
	
	-- timer
	
	self.spawnTimerId = nil
	self.spawnTime = 0
	
	-- auto allow...
	
	self.allowedVehicles = {}
	self.allowWalker = false
	self.allowPallets = getBoolFromUserAttribute(nodeId, "allowPallets", self.isLimitedObjectTypePallet)
	self.allowBales = getBoolFromUserAttribute(nodeId, "allowBales", self.isLimitedObjectTypeBale)
	self.allowObjects = not self.allowPallets and not self.allowBales
	-- self.allowWood ??
	
	self:addTrigger()
	
	-- end
	
	self:printFn('UPK_ObjectSpawner:new done')
	
	return self
end

function UPK_ObjectSpawner:fitCollisionMaskToAllowedVehicles()
	-- overwrite default function
	self:printFn('UPK_ObjectSpawner:fitCollisionMaskToAllowedVehicles()')
	
	local trigger_dynamic_objects = 16777216 -- bales, pallets, logs
	
	local collisionMask_old = getCollisionMask(self.triggerId)
	local collisionMask_new = collisionMask_old
	
	if (self.allowBales or self.allowPallets or self.allowObjects) and bitAND(collisionMask_new,trigger_dynamic_objects)==0 then
		self:printInfo('Warning: allowBales, allowPallets or allowObjects is set to true but collisionMask was not fitting (fixed)')
		collisionMask_new = collisionMask_new + trigger_dynamic_objects
	end
	
	-- result
	
	if collisionMask_new ~= collisionMask_old then
		self:printInfo('Warning: set collisionMask according to allowed vehicles to ',collisionMask_new,' (you may want to fix that)')
		setCollisionMask(self.triggerId,collisionMask_new)
	end
end

function UPK_ObjectSpawner:triggerCallback(triggerId, otherActorId, onEnter, onLeave, onStay, otherShapeId)
	-- overwrite default function
	self:printFn('UPK_ObjectSpawner:triggerCallback(',triggerId,', ',otherActorId,', ',onEnter,', ',onLeave,', ',onStay,', ',otherShapeId,')')
	
		local object = g_currentMission.nodeObjects[otherShapeId] or
						g_currentMission.nodeObjects[otherActorId]
		self:printInfo('object is ',object)
		
		if object~=nil then
			if self.vehicleNodesInTrigger[object]==nil then
				self.vehicleNodesInTrigger[object]={}
			end
			
			self:printInfo('allowPallets ',self.allowPallets)
			self:printInfo('is pallet ',getIsPallet(object))
			self:printInfo('allowBales ',self.allowBales)
			self:printInfo('is bale ',getIsBale(object))
			self:printInfo('allowObjects ',self.allowObjects)
			self:printInfo('is object ',(object.isa~=nil and object:isa(self.objectClass)))
			
			if (self.allowPallets and getIsPallet(object)) or
				(self.allowBales and getIsBale(object)) or
				(self.allowObjects and object.isa~=nil and object:isa(self.objectClass)) then
				self:printInfo('object allowed')
				if onEnter then
					self:printInfo('onEnter')
					if not isInTable(self.vehicleNodesInTrigger[object],otherShapeId) then
						table.insert(self.vehicleNodesInTrigger[object],otherShapeId)
						if length(self.vehicleNodesInTrigger[object])==1 then
							self:vehicleEntered(object) -- only for one node of the same object
						end
					end
				elseif onLeave then
					self:printInfo('onLeave')
					removeValueFromTable(self.vehicleNodesInTrigger[object],otherShapeId)
					if length(self.vehicleNodesInTrigger[object])==0 then
						self.vehicleNodesInTrigger[object]=nil
						self:vehicleLeft(object)
					end
				end
			end
		end
		--[[
		if self.allowWood then
			local splitType = SplitUtil.splitTypes[getSplitType(otherActorId)]
			self:printAll('splitType is ',splitType)
			if splitType~=nil and splitType.woodChipsPerLiter>0 then
				if onEnter then
					self:vehicleEntered(otherActorId,true)
				elseif onLeave then
					self:vehicleLeft(otherActorId,false)
				end
			end
		end
		]]--

end

function UPK_ObjectSpawner:delete()
	self:printFn('UPK_ObjectSpawner:delete()')
	UPK_ObjectSpawner:superClass().delete(self)
end

function UPK_ObjectSpawner:postLoad()
	self:printFn('UPK_ObjectSpawner:postLoad()')
	UPK_ObjectSpawner:superClass().postLoad(self)
	if self.spawnTimerId==nil then -- no timer running
		self:triggerUpdate(nil,nil)
	end
end

function UPK_ObjectSpawner:loadExtraNodes(xmlFile, key)
	self:printFn('UPK_ObjectSpawner:loadExtraNodes(',xmlFile,', ',key,')')
	local spawnTime = getXMLFloat(xmlFile, key .. "#spawnTime")
	--if spawnTime~=nil then
	--	self.spawnTimerId = reviveTimer(self.spawnTimerId, spawnTime, "spawnObject", self)
	--end
	self.nrObjectsNotSold = getXMLInt(xmlFile, key .. "#nrObjectsNotSold") or self.nrObjectsNotSold
	return true
end

function UPK_ObjectSpawner:getSaveExtraNodes(nodeIdent)
	self:printFn('UPK_ObjectSpawner:getSaveExtraNodes(',nodeIdent,')')
	local nodes=""
	if self.spawnTimerId~=nil then
		nodes = nodes .. ' spawnTime="'..tostring(self.spawnTime)..'"'
	end
	if self.nrObjectsNotSold~=0 then
		nodes = nodes .. ' nrObjectsNotSold="'..tostring(self.nrObjectsNotSold)..'"'
	end
	return nodes
end

function UPK_ObjectSpawner:setEnable(isEnabled,alreadySent)
	self:printFn('UPK_ObjectSpawner:setEnable(',isEnabled,', ',alreadySent,')')
	UPK_ObjectSpawner:superClass().setEnable(self,isEnabled,alreadySent)
	if self.isEnabled then
		self.spawnTimerId = nil
		self:triggerUpdate(nil,nil)
	else
		removeTimer(self.spawnTimerId)
		self.spawnTimerId = nil
	end
end

function UPK_ObjectSpawner:triggerUpdate(object,isInTrigger)
	self:printFn('UPK_ObjectSpawner:triggerUpdate(',object,', ',isInTrigger,')')
	if self.isServer then
		self:printInfo('vehiclesInTrigger ',self.vehiclesInTrigger,', maxObjects ',self.maxObjects)
		if self.vehiclesInTrigger < self.maxObjects then
			self:printInfo('want to spawn next object in ',self.spawnDelay)
			self.spawnTimerId = reviveTimer(self.spawnTimerId, self.spawnDelay, "spawnObject", self)
		else
			if self.spawnTimerId~=nil then
				removeTimer(self.spawnTimerId)
				self.spawnTimerId = nil
			end
		end
		if object~=nil and object.upkObjectId~=nil and isInTrigger==false then
			-- money
			self:printInfo('need to charge some money')
			local price = self.pricePerObject
			if object.fillLevel~=nil then
				price = price + object.fillLevel * self:getPricePerLiter(object.fillType)
			end
			g_currentMission:addMoneyChange(-price, self.moneyChangeId);
			g_currentMission:showMoneyChange(self.moneyChangeId, self.i18n["finance_"..self.statName])
			object.upkObjectId = nil
			self.nrObjectsNotSold = self.nrObjectsNotSold - 1
		end
	end
	if self.isServer then
		-- retag objects after loading savegame
		if object~=nil and self.nrObjectsNotSold>0 and object.upkObjectId==nil and isInTrigger==true then
			object.upkObjectId = UPK_ObjectSpawner.getNextObjectId()
			self.nrObjectsNotSold = self.nrObjectsNotSold - 1
		end
	end
end

UPK_ObjectSpawner.getPricePerLiter = UPK_FillTrigger.getPricePerLiter

-- like Basketball.createNode
function UPK_ObjectSpawner.createNode(i3dFilename, x, y, z, rx, ry, rz)
	printFn('UPK_ObjectSpawner.createNode(i3dFilename, x, y, z, rx, ry, rz)')
	local nodeRoot = loadI3DFile(i3dFilename)
	if nodeRoot==nil or nodeRoot==0 then
		return false
	end
	local nodeId = getChildAt(nodeRoot,0)
	link(getRootNode(), nodeId)
	delete(nodeRoot)
	setTranslation(nodeId, x, y, z)
	setRotation(nodeId, rx, ry, rz)
	return nodeId
end

function UPK_ObjectSpawner.checkCollisionMask(nodeId)
	printFn('UPK_ObjectSpawner.checkCollisionMask(nodeId)')
	if nodeId==nil or nodeId==0 then
		return
	end
	
	local trigger_dynamic_objects = 16777216 -- bales, pallets, logs
	
	local collisionMask_old = getCollisionMask(nodeId)
	local collisionMask_new = collisionMask_old
	
	if bitAND(collisionMask_new,trigger_dynamic_objects)==0 then
		printInfo('Warning: object needs to have a collision mask for dynamic objects (fixed)')
		collisionMask_new = collisionMask_new + trigger_dynamic_objects
	end
	
	if collisionMask_new ~= collisionMask_old then
		printInfo('Warning: set collisionMask of object to ',collisionMask_new,' (you may want to fix that)')
		setCollisionMask(nodeId,collisionMask_new)
	end
end

function UPK_ObjectSpawner:spawnObject()
	self:printFn('UPK_ObjectSpawner:spawnObject()')
	
	if not self.isEnabled then
		return false
	end
	
	if self.vehiclesInTrigger >= self.maxObjects then
		return false
	end

	if self.isLimitedObjectTypeBale and not g_currentMission:getCanAddLimitedObject(FSBaseMission.LIMITED_OBJECT_TYPE_BALE) then
		self:printWarning('cannot spawn bale - too many bales exist')
		return false
	elseif self.isLimitedObjectTypePallet and not g_currentMission:getCanAddLimitedObject(FSBaseMission.LIMITED_OBJECT_TYPE_PALLET) then
		self:printWarning('cannot spawn pallet - too many pallets exist')
		return false
	end

	local x, y, z = unpack(__c({getWorldTranslation(self.nodeId)}) + self.spawnPosition)
	local y_terrain = getTerrainHeightAtWorldPos(g_currentMission.terrainRootNode, x, 0, z) + 0.2
	y = mathmax(y, y_terrain)
	local rx, ry, rz = unpack((__c({getWorldRotation(self.nodeId)}) + self.spawnRotation*(2*math.pi)))
	
	local object = self.objectClass:new(self.isServer, self.isClient)
	self:printInfo('a')
	if object.load~=nil and (object:load(self.objectFilename, x, y, z, rx, ry, rz) or true) then
		self:printInfo('b')
		self:printAll('spawning object in ',x,', ',y,', ',z)
		--object:register()
		self:operateAction('OnObjectSpawned')
	elseif object.setNodeId~=nil then
		self:printInfo('c')
		local objectNodeId = UPK_ObjectSpawner.createNode(self.objectFilename, x, y, z, rx, ry, rz)
		if objectNodeId==false then
			self:printErr('couldnt load ',self.objectFilename)
			object:delete()
			return false
		end
		object:setNodeId(objectNodeId)
		
		-- taking care of saving objects
		
		self:operateAction('OnObjectSpawned')
	else
		self:printErr('couldnt load ',self.objectFilename)
		object:delete()
		return false
	end
	
	UPK_ObjectSpawner.checkCollisionMask(object.nodeId)
	
	-- tag object
	
	object.upkObjectId = UPK_ObjectSpawner.getNextObjectId()
	self.nrObjectsNotSold = self.nrObjectsNotSold + 1
	
	-- adjust fill level and type
	
	if self.objectFillType~=nil and object.fillType~=nil then
		if object.setFillType~=nil then
			object:setFillType(self.objectFillType)
		else
			-- object.fillType = self.objectFillType
		end
	end
	
	if self.objectFillLevel~=nil and object.fillLevel~=nil then
		if object.setFillLevel~=nil then
			object:setFillLevel(self.objectFillLevel)
		else
			-- object.fillLevel = self.objectFillLevel
		end
	end
	
	if self.objectCapacity~=nil and object.capacity~=nil then
		if object.setCapacity~=nil then
			object:setCapacity(self.objectCapacity)
		else
			-- object.capacity = self.objectCapacity
		end
	end

	-- what to sync?
	
	return true
end