-- by mor2000

--------------------
-- UPK_ObjectTrigger (reacts to objects like pallets, bales, bigbags, basketballs, .. you name it)

local UPK_ObjectTrigger_mt = ClassUPK(UPK_ObjectTrigger, UniversalProcessKit)
InitObjectClass(UPK_ObjectTrigger, "UPK_ObjectTrigger")
UniversalProcessKit.addModule("objecttrigger", UPK_ObjectTrigger)

function UPK_ObjectTrigger:new(nodeId, parent)
	printFn('UPK_ObjectTrigger:new(',nodeId,', ',parent,')')
	local self = UniversalProcessKit:new(nodeId, parent, UPK_ObjectTrigger_mt)
	registerObjectClassName(self, "UPK_ObjectTrigger")
	
	-- object
	
	-- just use attribute "object" with pre-set object types like "roundbaleGrass"
	-- or provide a filename in "object" and a class name in "objectClass" (default: Object)
	
	local objectStr = getStringFromUserAttribute(nodeId, "object", "")
	
	self.objectClass = nil
	
	if objectStr~="" and UPK_ObjectSpawner.objectTypes[objectStr]~=nil then
		-- preset object
		self:printInfo('preset object ',objectStr,' chosen')
		self.objectFilename = getLongFilename(UPK_ObjectSpawner.objectTypes[objectStr],nil)
		self.objectClass = UPK_ObjectSpawner.objectTypeToClass[objectStr]
	else
		-- only objectClass important
		local objectClassname = getStringFromUserAttribute(nodeId, "objectClass", "PhysicsObject")
			
		if _g[objectClassname]~=nil and type(_g[objectClassname])=="table" then
			self.objectClass = _g[objectClassname]
		elseif self.i18nNameSpace~=nil and _g[self.i18nNameSpace][objectClassname]~=nil and type(_g[self.i18nNameSpace][objectClassname])=="table" then
			self.objectClass = _g[self.i18nNameSpace][objectClassname]
		else
			self:printErr('no object class "'..objectClassname..'" found')
			return false
		end
	end
	
	if not (self.objectClass==PhysicsObject or derivesFromClass(self.objectClass,PhysicsObject)) then
		self:printErr('object to raect to is not an PhysicsObject')
		return false
	end
	
	self.isLimitedObjectTypeBale = (self.objectClass==Bale or derivesFromClass(self.objectClass,Bale))
	self.isLimitedObjectTypePallet = (self.objectClass==FillablePallet or derivesFromClass(self.objectClass,FillablePallet))

	-- delete objects
	
	self.deleteObjectOnEnter = getBoolFromUserAttribute(nodeId, "deleteObjectOnEnter", false)
	self.deleteObjectOnLeave = getBoolFromUserAttribute(nodeId, "deleteObjectOnLeave", false)

	-- auto allow...
	
	self.allowedVehicles = {}
	self.allowWalker = false
	self.allowPallets = getBoolFromUserAttribute(nodeId, "allowPallets", self.isLimitedObjectTypePallet)
	self.allowBales = getBoolFromUserAttribute(nodeId, "allowBales", self.isLimitedObjectTypeBale)
	self.allowObjects = getBoolFromUserAttribute(nodeId, "allowObjects", true)
	-- self.allowWood ??
	
	self:addTrigger()
	
	-- end
	
	self:printFn('UPK_ObjectTrigger:new done')
	
	return self
end

UPK_ObjectTrigger.fitCollisionMaskToAllowedVehicles = UPK_ObjectSpawner.fitCollisionMaskToAllowedVehicles

function UPK_ObjectTrigger:triggerCallback(triggerId, otherActorId, onEnter, onLeave, onStay, otherShapeId)
	-- overwrite default function
	self:printFn('UPK_ObjectTrigger:triggerCallback(',triggerId,', ',otherActorId,', ',onEnter,', ',onLeave,', ',onStay,', ',otherShapeId,')')
	if self.isEnabled then
		local object = g_currentMission.nodeObjects[otherShapeId] or
						g_currentMission.nodeObjects[otherActorId]
		self:printInfo('object is ',object)
		
		if object~=nil then
			if self.vehicleNodesInTrigger[object]==nil then
				self.vehicleNodesInTrigger[object]={}
			end
			
			self:printInfo('allowPallets ',self.allowPallets)
			self:printInfo('is pallet ',getIsPallet(object))
			self:printInfo('allowBales ',self.allowBales)
			self:printInfo('is bale ',getIsBale(object))
			self:printInfo('allowObjects ',self.allowObjects)
			self:printInfo('is object ',(object.isa~=nil and object:isa(self.objectClass)))
			
			if (self.allowPallets and getIsPallet(object)) or
				(self.allowBales and getIsBale(object)) or
				(self.allowObjects and object.isa~=nil and object:isa(self.objectClass)) then
				self:printInfo('object allowed')
				if onEnter then
					self:printInfo('onEnter')
					if not isInTable(self.vehicleNodesInTrigger[object],otherShapeId) then
						if self.deleteObjectOnEnter then
							object:delete()
						else
							table.insert(self.vehicleNodesInTrigger[object],otherShapeId)
							if length(self.vehicleNodesInTrigger[object])==1 then
								self:vehicleEntered(object) -- only for one node of the same object
							end
						end
					end
				elseif onLeave then
					self:printInfo('onLeave')
					removeValueFromTable(self.vehicleNodesInTrigger[object],otherShapeId)
					if length(self.vehicleNodesInTrigger[object])==0 then
						self.vehicleNodesInTrigger[object]=nil
						self:vehicleLeft(object)
						if self.deleteObjectOnLeave then
							object:delete()
						end
					end
				end
			end
		end
		--[[
		if self.allowWood then
			local splitType = SplitUtil.splitTypes[getSplitType(otherActorId)]
			self:printAll('splitType is ',splitType)
			if splitType~=nil and splitType.woodChipsPerLiter>0 then
				if onEnter then
					self:vehicleEntered(otherActorId,true)
				elseif onLeave then
					self:vehicleLeft(otherActorId,false)
				end
			end
		end
		]]--
	end
end

function UPK_ObjectTrigger:delete()
	self:printFn('UPK_ObjectTrigger:delete()')
	UPK_ObjectTrigger:superClass().delete(self)
end