-- by mor2000

--------------------
-- UPK_SprayerFillTrigger (fills trailers and/or shovels with specific fillType)

local UPK_SprayerFillTrigger_mt = ClassUPK(UPK_SprayerFillTrigger,UniversalProcessKit)
InitObjectClass(UPK_SprayerFillTrigger, "UPK_SprayerFillTrigger")
UniversalProcessKit.addModule("sprayerfilltrigger",UPK_SprayerFillTrigger)

function UPK_SprayerFillTrigger:new(nodeId, parent)
	printFn('UPK_SprayerFillTrigger:new(',nodeId,', ',parent,')')
	local self = UniversalProcessKit:new(nodeId, parent, UPK_SprayerFillTrigger_mt)
	registerObjectClassName(self, "UPK_SprayerFillTrigger")

	-- fillType
	self.fillFillType = UniversalProcessKit.FILLTYPE_FERTILIZER
	local fillFillTypeStr = getStringFromUserAttribute(nodeId, "fillType", "fertilizer")
	if UniversalProcessKit.fillTypeNameToInt[fillFillTypeStr]==UniversalProcessKit.FILLTYPE_LIQUIDFERTILIZER then
		self.fillFillType = UniversalProcessKit.FILLTYPE_LIQUIDFERTILIZER
	end
	
    self.createFillType = getBoolFromUserAttribute(nodeId, "createFillType", false)
    self.pricePerLiter = getNumberFromUserAttribute(nodeId, "pricePerLiter", 0)

	self.preferMapDefaultPrice = getBoolFromUserAttribute(nodeId, "preferMapDefaultPrice", false)
	self.pricePerLiterMultiplier = getVectorFromUserAttribute(nodeId, "pricePerLiterMultiplier", "1 1 1")
	self.pricesPerLiter = {}
	
	self.fillOnlyWholeNumbers = getBoolFromUserAttribute(nodeId, "fillOnlyWholeNumbers", false)
	self.amountToFillOfVehicle = {}
	
	-- statName
	
	self.statName=getStatNameFromUserAttribute(nodeId)

	self.allowedVehicles={}
	
	self.allowedVehicles[UniversalProcessKit.VEHICLE_SPRAYER] = getBoolFromUserAttribute(nodeId, "allowSprayer", true)
	
	self.allowWalker = false
	
	self:addTrigger()
	
	-- actions
	self:getActionUserAttributes('IfFilling')
	UniversalProcessKit.syncActionThruStream('IfFilling')
	
	self:getActionUserAttributes('IfFillingStarted')
	self:getActionUserAttributes('IfFillingStopped')
	
	self:printFn('UPK_SprayerFillTrigger:new done')
	
    return self
end

function UPK_SprayerFillTrigger:delete()
	self:printFn('UPK_SprayerFillTrigger:delete()')
	UPK_SprayerFillTrigger:superClass().delete(self)
end

function UPK_SprayerFillTrigger:triggerUpdate(vehicle,isInTrigger)
	self:printFn('UPK_SprayerFillTrigger:triggerUpdate(',vehicle,', ',isInTrigger,')')
	if self.isEnabled and self.isClient then
		if self.allowedVehicles[UniversalProcessKit.VEHICLE_SPRAYER] and UniversalProcessKit.isVehicleType(vehicle, UniversalProcessKit.VEHICLE_SPRAYER) then
			if not vehicle:allowFillType(self.fillFillType) then
				return
			end
			--self:print('recognized sprayer')
			if isInTrigger then
				--self:print('is in trigger')
				if vehicle.addFillTrigger~=nil and not vehicle.upk_sprayerFillTriggerAdded then
					self.amountToFillOfVehicle[vehicle]=0
					--self:print('adding trigger')
					vehicle:addFillTrigger(self)
					vehicle.upk_sprayerFillTriggerAdded = true
				end
			else
				if vehicle.removeFillTrigger~=nil then
					self.amountToFillOfVehicle[vehicle]=nil
					--self:print('removing trigger')
					vehicle:removeFillTrigger(self)
					vehicle.upk_sprayerFillTriggerAdded = false
				end
			end
		end
	end
end

UPK_SprayerFillTrigger.getFillLevel = UPK_FillTrigger.getFillLevel
UPK_SprayerFillTrigger.getPricePerLiter = UPK_FillTrigger.getPricePerLiter
UPK_SprayerFillTrigger.fill = UPK_FillTrigger.fillTrailer
UPK_SprayerFillTrigger.getIsActivatable = UPK_FillTrigger.getIsActivatable

